""""""
from sys import stdout
from typing import List, Dict, Optional, Callable, Tuple, Any
import logging
import functools
import json
import time
from itertools import repeat

from concurrent.futures import ThreadPoolExecutor
import requests as rq
from requests.exceptions import Timeout, ReadTimeout

import pandas as pd

from shimoku_api_python.api.plot_api import BasePlot, PlotApi


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logging.basicConfig(
    stream=stdout,
    datefmt='%Y-%m-%d %H:%M',
    format='%(asctime)s | %(levelname)s | %(message)s'
)


# TODO this should not go here but in a different script, maybe aux.py?
def with_retries(
        _func: Optional[Callable] = None,
        *, max_retries: int = 3, exponential_base: int = 2,
) -> Callable:
    """
    Decorate a function for retries.
    Exponential backoff.
    """

    def wrapper(func):
        @functools.wraps(func)
        def retries(*args, **kwargs):
            """
            Attempt a transaction with retries.
            """
            retry_exceptions = (
                'ProvisionedThroughputExceededException',
                'ThrottlingException'
            )

            retries = 0
            while retries < max_retries:
                pause_time = exponential_base ** retries
                # logging.info('Back-off set to %d seconds', pause_time)

                try:
                    return func(*args, **kwargs)
                # This one for Extractions from origin that return an XML / HTML rather than a proper response.
                except json.decoder.JSONDecodeError:
                    # For when the content cannot be decoded. Usually a bad response HTML/XML.
                    # logger.warning(f'Request response raised JSONDecodeError in {func.__name__} | '
                    #               f'Retry {retries} | Sleeping {exponential_base ** retries}. \n {e}')
                    time.sleep(pause_time)
                    retries += 1
                # Requests errors.
                except Timeout as e:
                    # logger.warning(f'Request Timeout Error in {func.__name__} | Retry {retries} | '
                    #               f'Sleeping {exponential_base ** retries}. \n {e}')
                    print('Timeout | Retrying')
                    time.sleep(pause_time)
                    retries += 1
                except ReadTimeout as e:
                    print('ReadTimeout | Retrying')
                    time.sleep(pause_time)
                    retries += 1
                except ConnectionError:
                    print('ConnectionError | Retrying')
                    time.sleep(pause_time)
                    retries += 1

            # raise MaxRetriesExceeded(f'Too many retries for {func.__name__} method')
            raise ValueError(f'Too many retries for {func.__name__} method')

        return retries

    return wrapper if not _func else wrapper(_func)


class AiPlotAdapter(BasePlot):
    _table = PlotApi.table


class AiAPI(AiPlotAdapter):

    def __init__(self, api_client, **kwargs):
        self.api_client = api_client

        if kwargs.get('business_id'):
            self.business_id: Optional[str] = kwargs['business_id']
        else:
            self.business_id: Optional[str] = None

    def predict_categorical(
            self, df_test: pd.DataFrame,
            model_endpoint: Optional[str] = None,
            app_name: Optional[str] = None, model_name: Optional[str] = None,
            explain: bool = False, append_test_data: bool = True,
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:

        def _create_request_data(window: int = 100):
            """Iterate over a dataframe in chunks of N row returning
            a list of json that can be sent to the ML API endpoint

            The size of the chunks is defined by `window`
            """
            raw_json: str = df_test.to_json(orient='records')
            raw_data: Dict = json.loads(raw_json)
            size = len(raw_data)
            for i in range(0, size, window):
                max_value = i + window
                if max_value > size:
                    max_value = size
                chunk_: List = raw_data[i: max_value]
                chunk_: List = [json.dumps([conditions, raw_datum]) for raw_datum in chunk_]
                yield chunk_

        def _clean_response(response_: Tuple, explainability_col_name: str = 'explainability') -> Dict:
            """Create the result out of the response

            Example
            ---------------
            input
                {'ID_POL': {'0': 651407},
                 'ID_CND': {'0': 1},
                 'PRIMER_EFECTO': {'0': '2018-01-13 00:00:00'},
                 'T_SEXO_TOMADOR': {'0': 'M'},
                 'prediction_probability': {'0': 0.6594810000302692},
                 'explainability': {
                    'sub_price_0': {'0': -0.18447917592951263},
                    'sub_npays_year': {'0': -0.0033977641276157094},
                    'hol_age': {'0': 0.0319888495770443},
                    'hol_seniority_driver': {'0': 0.016372795631977008}
                 },
                }

            output
                {'ID_POL': 653315,
                 'ID_CND': 1,
                 'PRIMER_EFECTO': '2018-01-25 12:14:00',
                 'T_SEXO_TOMADOR': 'F',
                 'prediction_probability': 0.7442640821276456,
                 'explainability': {
                    'sub_price_0': -0.32314667154583626,
                    'sub_npays_year': -0.018958441618165747,
                    'hol_age': 0.02434736937940205,
                    'hol_seniority_driver': 0.0008888141433092845,
                 }
                }
            """
            # position 0 contains the request status code
            if response_[0] < 200 or response_[0] > 300:  # position 1 contains the input
                d = json.loads(response_[1])[1]
                d['prediction_probability'] = 'ERROR'
                if explain:
                    d[explainability_col_name] = 'ERROR'
                return d

            # position 2 contains the request output
            result_: Dict = json.loads(response_[2])
            d: Dict = {k: v['0'] for k, v in result_.items() if k != explainability_col_name}
            if explain:
                d[explainability_col_name] = {
                    k: v['0'] for k, v in result_.get(explainability_col_name).items()
                }

            if not append_test_data:
                # TODO keep only: id, prediction column and explainability columns
                # cols_to_avoid = [c for c in df_test.columns if c != 'ID_POL']
                # d = {k: v for k, v in d.items() if k not in cols_to_avoid}
                pass

            return d

        @with_retries
        def _http_get_with_requests(url_: str, data: Dict) -> (int, Dict[str, Any], bytes):
            time.sleep(.25)
            response = rq.request('POST', url_, data=data)

            response_content = None
            try:
                response_content = response.content
            except Exception:
                pass

            return response.status_code, data, response_content

        def _http_get_with_requests_parallel(
                url: List[str], chunk_: List[Dict]
        ) -> Tuple[List, List]:
            rs: List = []
            rs_error: List = []
            executor = ThreadPoolExecutor(max_workers=100)
            for result in executor.map(_http_get_with_requests, repeat(url), chunk_):
                if result[0] >= 300:  # status_code
                    rs_error.append(result)
                else:
                    rs.append(result)

            # Another retry
            if rs_error:
                time.sleep(1)
                new_chunk = [element[1] for element in rs_error]
                rs_error = []
                for result in executor.map(_http_get_with_requests, repeat(url), new_chunk):
                    if result[0] >= 300:  # status_code
                        rs_error.append(result)
                    else:
                        rs.append(result)

            return rs, rs_error

        if not model_endpoint:
            if not app_name or not model_name:
                raise ValueError('model_endpoint OR app_name and model_name are required')

        results: List = []
        results_error: List = []
        # conditions: Dict = {"prediction_probability": True, "explainability": explain}
        conditions: Dict = {"churn_probability": True, "shap_contributions": explain}
        crd = _create_request_data(100)
        for chunk in crd:
            rs_, rs_error_ = _http_get_with_requests_parallel(url=model_endpoint, chunk_=chunk)
            results = results + [
                _clean_response(r, explainability_col_name='shap_contributions')
                for r in rs_
            ]
            results_error = results_error + [
                _clean_response(r, explainability_col_name='shap_contributions')
                for r in rs_error_
            ]

        return pd.DataFrame(results), pd.DataFrame(results_error)

    def train_model(self) -> str:
        """
        returns the model endpoint
        """
        raise NotImplementedError

    def predictive_table(
            self, df_test: pd.DataFrame,
            target_column: str, column_to_predict: str,
            menu_path: str, order: int,
            model_endpoint: Optional[str] = None,
            app_name: Optional[str] = None, model_name: Optional[str] = None,
            explain: bool = True,
            prediction_type: str = 'categorical',
            add_filter_by_column_to_predict: bool = True,
            add_search_by_target_column: bool = True,
            extra_filter_columns: Optional[List[str]] = None,
            extra_search_columns: Optional[List[str]] = None,
    ):
        """Predict a tabular dataset and store the results in a model

        Example
        --------------------
        target_column = 'customer_id'
        column_to_predict = 'y'

        :param df_test:
        :param target_column:
        :param column_to_predict:
        :param menu_path:
        :param order:
        :param model_endpoint:
        :param app_name:
        :param model_name:
        :param explain:
        :param prediction_type:
        :param add_filter_by_column_to_predict:
        :param add_search_by_target_column:
        :param extra_filter_columns:
        :param extra_search_columns:
        :return:
        """
        if not model_endpoint:
            if not app_name or not model_name:
                raise ValueError('model_endpoint OR app_name and model_name are required')

        available_predictions: List[str] = ['categorical']
        if prediction_type not in available_predictions:
            raise ValueError(
                f'{prediction_type} not allowed | '
                f'Prediction type available are {available_predictions}'
            )

        df_result, _ = self.predict_categorical(
            df_test=df_test, model_endpoint=model_endpoint, explain=explain,
        )

        if add_filter_by_column_to_predict:
            filter_columns: List[str] = [column_to_predict]
        else:
            filter_columns = []

        if add_search_by_target_column:
            search_columns: List[str] = [target_column]
        else:
            search_columns = []

        if extra_filter_columns:
            filter_columns = filter_columns + extra_filter_columns
        if extra_search_columns:
            search_columns = search_columns + extra_search_columns

        self._table(
            data=df_result,
            menu_path=menu_path,
            order=order,
            filter_columns=filter_columns,
            search_columns=search_columns,
        )



""""""

from typing import List, Dict, Optional, Any, Union
import json
from time import sleep

from shimoku_api_python.exceptions import ApiClientError


class GetExplorerAPI(object):

    def __init__(self, api_client):
        self.api_client = api_client

    def get_business(self, business_id: str, **kwargs) -> Dict:
        """Retrieve an specific user_id

        :param business_id: user UUID
        """
        endpoint: str = f'business/{business_id}'
        business_data: Dict = (
            self.api_client.query_element(
                method='GET', endpoint=endpoint, **kwargs
            )
        )
        return business_data

    def get_app_type(self, app_type_id: str, **kwargs) -> Dict:
        """Retrieve an specific app_id metadata

        :param app_type_id: app type UUID
        """
        endpoint: str = f'apptype/{app_type_id}'
        app_type_data: Dict = (
            self.api_client.query_element(
                method='GET', endpoint=endpoint, **kwargs
            )
        )
        return app_type_data

    def get_app(self, business_id: str, app_id: str, **kwargs) -> Dict:
        """Retrieve an specific app_id metadata

        :param business_id: business UUID
        :param app_id: app UUID
        """
        endpoint: str = f'business/{business_id}/app/{app_id}'
        app_data: Dict = (
            self.api_client.query_element(
                method='GET', endpoint=endpoint, **kwargs
            )
        )
        return app_data

# TODO add new data & dataset logic!
    def _get_report_with_data(
        self,
        business_id: Optional[str] = None,
        app_id: Optional[str] = None,
        report_id: Optional[str] = None,
        external_id: Optional[str] = None,
        **kwargs,
    ) -> Dict:
        """Retrieve an specific report data

        :param business_id: business UUID
        :param app_id: Shimoku app UUID (only required if the external_id is provided)
        :param report_id: Shimoku report UUID
        :param external_id: external report UUID
        """
        if report_id:
            endpoint: str = f'business/{business_id}/app/{app_id}/report/{report_id}'
            report_data: Dict = (
                self.api_client.query_element(
                    method='GET',
                    endpoint=endpoint,
                    **kwargs
                )
            )
        elif external_id:
            if not app_id:
                raise ValueError(
                    'If you retrieve by external_id '
                    'you must provide an app_id'
                )

            report_ids_in_app: List[str] = (
                self.get_app_all_reports(app_id)
            )

            for report_id in report_ids_in_app:
                report_data_: Dict = self.get_report(report_id=report_id)
                if report_data_['etl_code_id'] == external_id:
                    endpoint: str = (
                        f'business/{business_id}/'
                        f'app/{app_id}/'
                        f'report/{report_id}'
                    )
                    report_data: Dict = (
                        self.api_client.query_element(
                            method='GET', endpoint=endpoint, **kwargs
                        )
                    )
                    return report_data
            else:
                return {}
        else:
            raise ValueError('Either report_id or external_id must be provided')

        if report_data.get('chartData'):
            report_data['chartData'] = json.loads(report_data['chartData'])
        return report_data

    def get_report(
        self,
        business_id: Optional[str] = None,
        app_id: Optional[str] = None,
        report_id: Optional[str] = None,
        external_id: Optional[str] = None,
        **kwargs,
    ) -> Dict:
        """Retrieve an specific report data

        :param business_id: business UUID
        :param app_id: Shimoku app UUID (only required if the external_id is provided)
        :param report_id: Shimoku report UUID
        :param external_id: external report UUID
        """
        report_data: Dict = (
            self._get_report_with_data(
                business_id=business_id,
                app_id=app_id,
                report_id=report_id,
                external_id=external_id,
            )
        )
        # we do not return the chartData in the get_report()
        #  use _get_report_with_data() instead
        if report_data.get('chartData'):
            report_data.pop('chartData')
        return report_data

    def get_dataset(self, business_id: str, dataset_id: str, **kwargs) -> Dict:
        """Retrieve an specific app_id metadata

        :param business_id: business UUID
        :param dataset_id: dataset UUID
        """
        endpoint: str = f'business/{business_id}/dataset/{dataset_id}'
        dataset_data: Dict = (
            self.api_client.query_element(
                method='GET', endpoint=endpoint, **kwargs
            )
        )
        return dataset_data

    def get_reportdataset(
            self, business_id: str, app_id: str, report_id: str,
            reportdataset_id: str, **kwargs
    ) -> Dict:
        """Retrieve an specific app_id metadata

        :param business_id: business UUID
        :param app_id: app UUID
        :param report_id: report UUID
        :param reportdataset_id: reportDataSet UUID
        """
        endpoint: str = (
            f'business/{business_id}/'
            f'app/{app_id}/'
            f'report/{report_id}/'
            f'{reportdataset_id}'
        )
        dataset_data: Dict = (
            self.api_client.query_element(
                method='GET', endpoint=endpoint, **kwargs
            )
        )
        return dataset_data

# TODO add new data & dataset logic!
    def get_report_data(
        self, business_id: str,
        app_id: Optional[str] = None,
        report_id: Optional[str] = None,
        external_id: Optional[str] = None,
    ) -> List[Dict]:
        """"""
        report: Dict = self.get_report(
            business_id=business_id,
            app_id=app_id,
            report_id=report_id,
        )

        if report['reportType']:
            report: Dict = (
                self._get_report_with_data(
                    business_id=business_id,
                    app_id=app_id,
                    report_id=report_id,
                    external_id=external_id,
                )
            )
            report_data: List = report.get('chartData')
            if report_data:
                return report_data
            else:
                return list()
        else:
            endpoint: str = (
                f'business/{business_id}/'
                f'app/{app_id}/'
                f'report/{report_id}/reportEntries'
            )
            report_entries: List = [
                self.api_client.query_element(
                    method='GET', endpoint=endpoint,
                )
            ]
            return report_entries[0]['items']

    def get_file(
            self, business_id: Optional[str] = None,
            app_id: Optional[str] = None,
            file_id: Optional[str] = None,
    ) -> bytes:
        """Retrieve an specific file from an app

        :param business_id: business UUID
        :param app_id: Shimoku app UUID (only required if the external_id is provided)
        :param file_id: Shimoku report UUID
        """
        endpoint: str = f'business/{business_id}/app/{app_id}/file/{file_id}'
        file_data: str = self.api_client.query_element(method='GET', endpoint=endpoint)

        try:
            url: str = file_data['url']
        except KeyError:
            raise KeyError(f'Could not GET file')

        file_object = self.api_client.raw_request(
            **dict(method='GET', url=url)
        )
        return file_object.content

    def get_files(
            self, business_id: Optional[str] = None,
            app_id: Optional[str] = None,
    ) -> List[Dict]:
        """Retrieve an specific file from an app

        :param business_id: business UUID
        :param app_id: Shimoku app UUID (only required if the external_id is provided)
        """
        endpoint: str = f'business/{business_id}/app/{app_id}/files'
        files: List[Dict] = self.api_client.query_element(method='GET', endpoint=endpoint)
        return files


class CascadeExplorerAPI(GetExplorerAPI):

    def __init__(self, api_client):
        super().__init__(api_client)

    def get_universe_businesses(self) -> List[Dict]:
        endpoint: str = f'businesses'
        return (
            self.api_client.query_element(
                endpoint=endpoint, method='GET',
            )
        )['items']

    def find_business_by_name_filter(
        self, name: Optional[str] = None,
    ) -> Dict:
        """"""
        businesses: List[Dict] = self.get_universe_businesses()

        businesses: List[Dict] = [
            business
            for business in businesses
            if business['name'] == name
        ]
        if not businesses:
            return {}

        assert len(businesses) == 1
        business: Dict = businesses[0]
        return business

    def get_universe_app_types(self) -> List[Dict]:
        endpoint: str = f'apptypes'
        return (
            self.api_client.query_element(
                endpoint=endpoint, method='GET',
            )
        )['items']

    def find_app_type_by_name_filter(
        self, name: Optional[str] = None,
        normalized_name: Optional[str] = None,
    ) -> Dict:
        """"""
        app_types: List[Dict] = self.get_universe_app_types()

        if name:
            app_types: List[Dict] = [
                app_type
                for app_type in app_types
                if app_type['name'] == name
            ]
        elif normalized_name:
            app_types: List[Dict] = [
                app_type
                for app_type in app_types
                if app_type['normalizedName'] == normalized_name
            ]

        if not app_types:
            return {}

        assert len(app_types) == 1
        app_type: Dict = app_types[0]
        return app_type

    def get_business_apps(self, business_id: str) -> List[Dict]:
        """Given a business retrieve all app metadata

        :param business_id: business UUID
        """
        endpoint: str = f'business/{business_id}/apps'
        apps_raw: Dict = (
            self.api_client.query_element(
                endpoint=endpoint, method='GET',
            )
        )
        apps: List[Dict] = apps_raw.get('items')

        if not apps:
            return []
        return apps

    def get_business_app_ids(self, business_id: str) -> List[str]:
        """Given a business retrieve all app ids

        :param business_id: business UUID
        """
        apps: Optional[List[Dict]] = (
            self.get_business_apps(
                business_id=business_id,
            )
        )
        return [app['id'] for app in apps]

    def get_business_all_files(self, business_id) -> List[Dict]:
        """Given a business retrieve all files metadata
        """
        apps: List[Dict] = self.get_business_apps(business_id=business_id)
        files: List[Dict] = list
        for app in apps:
            files = files + self.get_files(business_id=business_id, app_id=app['id'])
        return files

    def find_app_by_name_filter(
        self, business_id: str, name: Optional[str] = None,
        normalized_name: Optional[str] = None,
    ) -> Dict:
        """"""
        apps_list: List[Dict] = self.get_business_apps(business_id=business_id)

        if name:
            apps: List[Dict] = [
                app
                for app in apps_list
                if app['name'] == name
            ]
        elif normalized_name:
            apps: List[Dict] = [
                app
                for app in apps_list
                if app['normalizedName'] == normalized_name
            ]

        if not apps:
            return {}

        assert len(apps) == 1
        apps: Dict = apps[0]
        return apps

    def get_app_path_names(self, business_id: str, app_id: str) -> List[str]:
        """Given a Path that belongs to an AppId retrieve all reportId

        :param business_id: business UUID
        :param app_id: app UUID
        """
        reports: List[Dict] = (
            self.get_app_reports(
                business_id=business_id,
                app_id=app_id,
            )
        )

        paths: List[str] = []
        for report in reports:
            path: Optional[str] = report.get('path')
            if path:
                paths = paths + [path]
        return paths

    def get_app_reports(self, business_id: str, app_id: str) -> List[Dict]:
        """Given an App Id retrieve all reports data from all reports
        that belongs to such App Id.
        """
        endpoint: str = f'business/{business_id}/app/{app_id}/reports'
        reports_raw: Dict = (
            self.api_client.query_element(
                endpoint=endpoint, method='GET',
            )
        )
        reports = reports_raw.get('items')
        if not reports:
            return []
        return reports

    def get_app_report_ids(self, business_id: str, app_id: str) -> List[str]:
        """Given an app retrieve all report_id

        :param business_id: business UUID
        :param app_id: app UUID
        """
        reports: List[Dict] = (
            self.get_app_reports(
                business_id=business_id,
                app_id=app_id,
            )
        )
        return [report['id'] for report in reports]

    def get_report_datasets(
            self, business_id: str, app_id: str,  report_id: str,
    ) -> List[Dict]:
        endpoint: str = (
            f'business/{business_id}/'
            f'app/{app_id}/'
            f'report/{report_id}/'
            f'reportDataSets'
        )
        reportdatasets: List[Dict] = (
            self.api_client.query_element(
                endpoint=endpoint, method='GET',
            )
        )
        reportdatasets = reportdatasets.get('items')
        if not reportdatasets:
            return []

        data_sets: List[Dict] = []
        for reportdataset in reportdatasets:
            endpoint: str = (
                f'business/{business_id}/'
                f'dataSet/{reportdataset["dataSetId"]}'
            )
            data_set: Dict = (
                self.api_client.query_element(
                    endpoint=endpoint, method='GET',
                )
            )
            if data_set.get('reportDataSets'):
                data_sets = data_sets + [data_set]
        return data_sets

    def get_dataset_data(
            self, business_id: str, dataset_id: str,
    ) -> List[Dict]:
        endpoint: str = (
            f'business/{business_id}/'
            f'dataSet/{dataset_id}/'
            f'datas'
        )
        datas_raw: List[Dict] = (
            self.api_client.query_element(
                endpoint=endpoint, method='GET',
            )
        )
        datas = datas_raw.get('items')
        if not datas:
            return []
        return datas

    def get_report_dataset_data(
            self, business_id: str, app_id: str, report_id: str,
    ) -> List[Dict]:
        """"""
        report_datasets: List[Dict] = self.get_report_datasets(
            business_id=business_id, app_id=app_id, report_id=report_id,
        )
        data: List = []
        for report_dataset in report_datasets:
            dataset_id: str = report_dataset['id']
            datum: List[Dict] = (
                self.get_dataset_data(
                    business_id=business_id,
                    dataset_id=dataset_id,
                )
            )
            data = data + datum
        return data

    # TODO pending
    def get_report_all_report_entries(self, report_id: str) -> List[str]:
        """Given a report retrieve all reportEntries

        :param report_id: app UUID
        """
        raise NotImplementedError

    def get_path_report_ids(
        self, business_id: str, app_id: str, path_name: str,
    ) -> List[str]:
        """Given an App return all Reports ids that belong to a target path"""
        reports: List[Dict] = self.get_app_reports(
            business_id=business_id, app_id=app_id,
        )

        path_report_ids: List[str] = []
        for report in reports:
            path: Optional[str] = report.get('path')
            if path == path_name:
                report_id: str = report['id']
                path_report_ids = path_report_ids + [report_id]
        return path_report_ids

    def get_path_reports(
        self, business_id: str, app_id: str, path_name: str,
    ) -> List[Dict]:
        """Given an App return all Reports data that belong to a target path"""
        reports: List[Dict] = self.get_app_reports(
            business_id=business_id, app_id=app_id,
        )

        path_reports: List[Dict] = []
        for report in reports:
            path: Optional[str] = report.get('path')
            if path == path_name:
                path_reports = path_reports + [report]
        return path_reports

    def get_business_apps_with_filter(
            self, business_id: str, app_filter: Dict
    ) -> List[Dict]:
        """
        """
        apps: List[Dict] = (
            self.get_business_apps(
                business_id=business_id,
            )
        )

        apps: List[Dict] = []
        for app in apps:
            for filter_key, filter_value in app_filter.items():
                if app[filter_key] == filter_value:
                    apps.append(app)
        return apps

    def get_app_reports_by_filter(
        self, app_id: str,
        report_filter: Dict
    ) -> List[Dict]:
        """Having an AppId first retrieve all reportId that belongs
        to the target AppId. Second filter and take the reportId

        # TODO filter example!!
        """
        report_ids: List[str] = (
            self.get_app_all_reports(
                app_id=app_id,
            )
        )

        reports: List[Dict] = []
        for report_id in report_ids:
            report: Dict = self.get_report(report_id=report_id)
            for filter_key, filter_value in report_filter.items():
                if report[filter_key] == filter_value:
                    reports.append(report)
            return reports

    def get_app_type_by_name(
            self, name: Optional[str] = None,
            normalized_name: Optional[str] = None,
    ) -> Dict:
        """
        :param business_id: business UUID
        :param name: appType name
        :param normalized_name: appType normalizedName
        """
        if not name and not normalized_name:
            raise ValueError('You must provide either "name" or "normalized_name"')

        app_types: List[Dict] = self.get_universe_app_types()

        result: List[Dict] = [
            app_type
            for app_type in app_types
            if (
                    app_type['name'] == name
                    or app_type['normalizedName'] == normalized_name
            )
        ]

        if result:
            assert len(result) == 1
            return result[0]
        else:
            return {}

    def get_app_by_type(
        self, business_id: str, app_type_id: str,
    ) -> Dict:
        """
        :param business_id: business UUID
        :param app_type_id: appType UUID
        """
        apps: List[Dict] = self.get_business_apps(business_id=business_id)

        # Is expected to be a single item (Dict) but an App
        # could have several reports with the same name
        result: Any = {}
        for app in apps:
            if not app.get('type'):  # is not mandatory for app to have an AppType
                continue

            if app['type']['id'] == app_type_id:
                if result:
                    if len(result) == 1:
                        result: List[Dict] = result + [app]
                    else:
                        result: List[Dict] = result + [app]
                else:
                    result: List[Dict] = [app]
        if result:
            assert len(result) == 1
            return result[0]
        else:
            return {}

    def get_app_by_name(self, business_id: str, name: str) -> Dict:
        """
        :param business_id: business UUID
        :param name: app or apptype name
        """
        apps: List[Dict] = self.get_business_apps(business_id=business_id)

        # Is expected to be a single item (Dict) but an App
        # could have several reports with the same name
        result: Any = {}
        for app in apps:
            # if App name does not match check the AppType,
            # if it does not match the AppType Name then pass to the following App
            if app.get('name'):
                if not app['name'] == name:
                    continue
            else:
                if not app.get('type'):
                    continue
                try:
                    app_type: Dict = self.get_app_type(
                        app_type_id=app['type']['id'],
                    )
                except ApiClientError:  # Business admin user
                    continue

                if (
                    not app_type['normalizedName'] == name
                    and
                    not app_type['name'] == name
                ):
                    continue

            if result:
                if len(result) == 1:
                    result: List[Dict] = result + [app]
                else:
                    result: List[Dict] = result + [app]
            else:
                result: List[Dict] = [app]

        if result:
            assert len(result) == 1
            return result[0]
        else:
            return {}

    def get_app_by_url(self, business_id: str, url: str) -> Dict:
        """Internally known as normalizedName
        :param business_id: business UUID
        :param url: app url
        """
        apps: List[Dict] = self.get_business_apps(business_id=business_id)

        # Is expected to be a single item (Dict) but an App
        # could have several reports with the same name
        result: Any = {}
        for app in apps:
            # if App normalizedName does not match check the AppType,
            # if it does not match the AppType normalizedName then
            # pass to the following App
            if app.get('normalizedName'):
                if not app['normalizedName'] == url:
                    continue
            else:
                if not app.get('type'):
                    continue
                try:
                    app_type: Dict = self.get_app_type(
                        app_type_id=app['type']['id'],
                    )
                except ApiClientError:  # Business admin user
                    continue

                if (
                    not app_type['normalizedName'] == url
                    and
                    not app_type['name'] == url
                ):
                    continue

            if result:
                if len(result) == 1:
                    result: List[Dict] = result + [app]
                else:
                    result: List[Dict] = result + [app]
            else:
                result: List[Dict] = [app]

        if result:
            assert len(result) == 1
            return result[0]
        else:
            return {}


class CreateExplorerAPI(object):
    _find_business_by_name_filter = CascadeExplorerAPI.find_business_by_name_filter
    _find_app_type_by_name_filter = CascadeExplorerAPI.find_app_type_by_name_filter

    def __init__(self, api_client):
        self.api_client = api_client

    def _create_normalized_name(self, name: str) -> str:
        """Having a name create a normalizedName

        Example
        ----------------------
        # "name": "Test Borrar"
        # "normalizedName": "test-borrar"
        """
        return '-'.join(name.split(' ')).lower()

    def _create_key_name(self, name: str) -> str:
        """Having a name create a key

        Example
        ----------------------
        # "name": "Test Borrar"
        # "key": "TEST_BORRAR"
        """
        return '_'.join(name.split(' ')).upper()

    def create_business(self, name: str) -> Dict:
        """"""
        business: Dict = self._find_business_by_name_filter(name=name)
        if business:
            raise ValueError(f'A Business with the name {name} already exists')

        endpoint: str = 'business'

        item: Dict = {'name': name}

        return self.api_client.query_element(
            method='POST', endpoint=endpoint, **{'body_params': item},
        )

    def create_app_type(self, name: str) -> Dict:
        """"""
        app_type: Dict = self._find_app_type_by_name_filter(name=name)
        if app_type:
            raise ValueError(f'An AppType with the name {name} already exists')

        endpoint: str = 'apptype'
        # for instance:
        # "name": "Test Borrar"
        # "key": "TEST_BORRAR"
        # "normalizedName": "test-borrar"
        normalized_name: str = self._create_normalized_name(name)
        key: str = self._create_key_name(name)

        item: Dict = {
            'name': name,
            'key': key,
            'normalizedName': normalized_name,
        }

        return self.api_client.query_element(
            method='POST', endpoint=endpoint, **{'body_params': item},
        )

    def create_app(
        self, business_id: str,
        name: Optional[str],
        app_type_id: Optional[str] = None,
        app_metadata: Optional[Dict] = None,
    ) -> Dict:
        """
        """
        endpoint: str = f'business/{business_id}/app'
        item: Dict = {}

        if app_type_id:
            item['appTypeId'] = app_type_id

        normalized_name: str = self._create_normalized_name(name)
        item['name'] = name
        item['normalizedName'] = normalized_name

        if app_metadata:
            hide_title: bool = app_metadata.get('hideTitle')
            if hide_title:
                item['hideTitle'] = 'true' if hide_title else 'false'
            else:
                item['hideTitle'] = 'true'

            # These are the optional fields (previous were the mandatory ones)
            allowed_columns: List[str] = [
                'paymentType', 'trialDays',
                'appSubscriptionInUserId',
            ]
            # Check all kwargs keys are in the allowed_columns list
            assert all([key in allowed_columns for key in app_metadata.keys()])
            # Update items with kwargs
            item.update(app_metadata)
        else:
            item['hideTitle'] = 'true'

        return self.api_client.query_element(
            method='POST', endpoint=endpoint, **{'body_params': item},
        )

    def create_report(
        self, business_id: str, app_id: str, report_metadata: Dict,
        real_time: bool = False,
    ) -> Dict:
        """Create new Report associated to an AppId

        :param business_id:
        :param app_id:
        :param report_metadata: A dict with all the values required to create a report
        :param real_time: Whether it is real time or not
        """
        def append_fields(item: Dict, field_name: str) -> Dict:
            """Equivalent to
            grid: Optional[str] = report_metadata.get('grid')
            if grid:
                item['grid'] = grid
            """
            field_value: Optional[str] = report_metadata.get(field_name)
            if field_value is not None:
                item[field_name] = field_value
            return item

        endpoint: str = f'business/{business_id}/app/{app_id}/report'

        # These are the mandatory fields
        # title: str = report_metadata['title']

        # These are the mandatory fields
        item: Dict = {'appId': app_id}

        item: Dict = append_fields(item=item, field_name='title')
        item: Dict = append_fields(item=item, field_name='path')
        item: Dict = append_fields(item=item, field_name='pathOrder')
        item: Dict = append_fields(item=item, field_name='grid')
        item: Dict = append_fields(item=item, field_name='reportType')
        item: Dict = append_fields(item=item, field_name='order')
        item: Dict = append_fields(item=item, field_name='sizeColumns')
        item: Dict = append_fields(item=item, field_name='sizeRows')
        item: Dict = append_fields(item=item, field_name='padding')
        item: Dict = append_fields(item=item, field_name='bentobox')
        item: Dict = append_fields(item=item, field_name='properties')

        if real_time:
            item['subscribe'] = True

        # Update items with kwargs
        item.update(report_metadata)

        # Optional values
        report_type: str = report_metadata.get('reportType')
        if report_type:
            if report_type != 'Table':  # Tables have reportType as None
                item['reportType'] = report_type
            elif report_metadata.get('smartFilters'):
                # Smart filters only exists for Tables
                item['smartFilters'] = report_metadata['smartFilters']

        report: Dict = (
            self.api_client.query_element(
                method='POST', endpoint=endpoint,
                **{'body_params': item},
            )
        )

        return {
            k: v
            for k, v in report.items()
            if k not in ['chartData', 'owner', 'chartDataItem']  # we do not return the data
        }

    def create_dataset(self, business_id: str) -> Dict:
        """Create new DataSet associated to a business

        :param business_id:
        """
        endpoint: str = f'business/{business_id}/dataSet'

        return self.api_client.query_element(
            method='POST', endpoint=endpoint, **{'body_params': {}},
        )

    def create_reportdataset(
            self, business_id: str, app_id: str, report_id: str,
            dataset_id: str, dataset_properties: str,
    ) -> Dict:
        """Create new dataset associated to a Report

        :param business_id:
        :param app_id:
        :param report_id:
        :param dataset_id:
        :param dataset_properties: a json with the properties
        """
        endpoint: str = (
            f'business/{business_id}/'
            f'app/{app_id}/'
            f'report/{report_id}/'
            f'reportDataSet'
        )

        item: Dict = {
            'reportId': report_id,
            'dataSetId': dataset_id,
        }

        if dataset_properties:
            item['properties'] = dataset_properties

        return self.api_client.query_element(
            method='POST', endpoint=endpoint, **{'body_params': item},
        )

    def create_data_points(
            self, business_id: str, dataset_id: str,
            items: List[str],
    ) -> List[Dict]:
        """Create new row in Data (equivalent to reportEntry)

        :param business_id:
        :param dataset_id:
        :param items:
        """
        endpoint: str = (
            f'business/{business_id}/'
            f'dataSet/{dataset_id}/'
            f'data'
        )

        data: List[Dict] = []
        for item in items:
            datum: Dict = (
                self.api_client.query_element(
                    method='POST', endpoint=endpoint,
                    **{'body_params': item},
                )
            )
            sleep(.25)
            data = data + [datum]

        return data

    def _create_report_entries(
        self, business_id: str, app_id: str, report_id: str,
        items: List[Dict], batch_size: int = 999,
    ) -> List[Dict]:
        """Create new reportEntry associated to a Report

        :param business_id:
        :param app_id:
        :param report_id:
        :param items: A dict with all the values required to create a reportEntry
        """
        if batch_size >= 1000:
            raise ValueError('batch_size must be less than 1000')

        endpoint: str = (
            f'business/{business_id}/'
            f'app/{app_id}/'
            f'report/{report_id}/'
            f'reportEntry/batch'
        )

        #report_entries: List[Dict] = []
        for chunk in range(0, len(items), batch_size):
            #report_entries_batch = report_entries + (
            self.api_client.query_element(
                method='POST', endpoint=endpoint,
                **{'body_params': items[chunk:chunk + batch_size]},
            )
            #)
            sleep(.25)
            #report_entries = report_entries + [report_entries_batch]

        #return report_entries

    def create_file(
            self, business_id: str, app_id: str,
            file_metadata: Dict, file_object: bytes,
    ) -> Dict:
        """Create new Files associated to an AppId

        Example
        ------------
            file_metadata= {
                name: String
                fileName: String (It should be normalized)
                contentType: String (Content type of the file you want to upload. Ex: image/png)
            }

        :param business_id:
        :param app_id:
        :param file_metadata:
        """
        endpoint: str = f'business/{business_id}/app/{app_id}/file'

        file_data: str = (
            self.api_client.query_element(
                method='POST', endpoint=endpoint,
                **{'body_params': file_metadata},
            )
        )
        try:
            url: str = file_data['url']
        except KeyError:
            raise KeyError(f'Could not POST file')

        r = self.api_client.raw_request(
            **dict(
                method='PUT', url=url, data=file_object,
                headers={'Content-Type': 'text/csv'},
            )
        )
        try:
            assert all([r.status_code >= 200,  r.status_code < 300])
            file_data['Success'] = True
        except AssertionError:
            file_data['Success'] = False
        file_data['Status'] = r.status_code
        file_data.pop('url')
        return file_data


class UpdateExplorerAPI(CascadeExplorerAPI):
    _find_business_by_name_filter = CascadeExplorerAPI.find_business_by_name_filter
    _find_app_type_by_name_filter = CascadeExplorerAPI.find_app_type_by_name_filter

    def __init__(self, api_client):
        self.api_client = api_client

    def update_business(self, business_id: str, business_data: Dict) -> Dict:
        """"""
        name = business_data.get('name')
        if name:
            business: Dict = self._find_business_by_name_filter(name=name)
            if business:
                raise ValueError(
                    f'Cannot Update | '
                    f'A Business with the name {name} already exists'
                )

        endpoint: str = f'business/{business_id}'
        return self.api_client.query_element(
            method='PATCH', endpoint=endpoint, **{'body_params': business_data},
        )

    def update_app_type(self, app_type_id: str, app_type_metadata: Dict) -> Dict:
        """"""
        name = app_type_metadata.get('name')
        if name:
            _app_type: Dict = self._find_app_type_by_name_filter(name=name)
            if _app_type:
                raise ValueError(
                    f'Cannot Update | '
                    f'A AppType with the name {name} already exists'
                )

        endpoint: str = f'apptype/{app_type_id}'
        return self.api_client.query_element(
            method='PATCH', endpoint=endpoint, **{'body_params': app_type_metadata},
        )

    def update_app(self, business_id: str, app_id: str, app_metadata: Dict) -> Dict:
        """
        :param business_id:
        :param app_id:
        :param app_metadata: contain the elements to update key
            is the col name and value the value to overwrite
        """
        endpoint: str = f'business/{business_id}/app/{app_id}'
        return self.api_client.query_element(
            method='PATCH', endpoint=endpoint,
            **{'body_params': app_metadata},
        )

    def update_report(
            self, business_id: str, app_id: str, report_id: str,
            report_metadata: Dict,
    ) -> Dict:
        """"""
        endpoint: str = f'business/{business_id}/app/{app_id}/report/{report_id}'
        return self.api_client.query_element(
            method='PATCH', endpoint=endpoint,
            **{'body_params': report_metadata},
        )

    def update_reportdataset(
            self, business_id: str, app_id: str, report_id: str,
            reportdataset_id: str, reportdataset_metadata: Dict,
    ) -> Dict:
        """"""
        endpoint: str = (
            f'business/{business_id}/'
            f'app/{app_id}/'
            f'report/{report_id}/'
            f'{reportdataset_id}'
        )
        return self.api_client.query_element(
            method='PATCH', endpoint=endpoint,
            **{'body_params': reportdataset_metadata},
        )

    def update_dataset(
            self, business_id: str, dataset_id: str,
            dataset_metadata: Dict,
    ) -> Dict:
        """"""
        endpoint: str = f'business/{business_id}/dataset/{dataset_id}'
        return self.api_client.query_element(
            method='PATCH', endpoint=endpoint,
            **{'body_params': dataset_metadata},
        )


class MultiCascadeExplorerAPI(CascadeExplorerAPI):

    def __init__(self, api_client):
        super().__init__(api_client)

    # TODO paginate
    def get_business_paths(self, business_id: str) -> List[str]:
        """Given a business retrieve all path names

        :param business_id: business UUID
        """
        apps: List[Dict] = self.get_business_apps(business_id=business_id)
        paths: List[str] = []
        for app in apps:
            app_id: str = app['id']
            app_paths: List[str] = self.get_app_paths(app_id=app_id)
            paths = paths + app_paths
        return paths

    # TODO paginate
    def get_business_reports(self, business_id: str) -> List[str]:
        """Given a business retrieve all report ids

        :param business_id: business UUID
        """
        apps: List[Dict] = self.get_business_apps(business_id=business_id)
        report_ids: List[str] = []
        for app in apps:
            app_id: str = app['id']
            app_reports: List[Dict] = self.get_app_reports(
                business_id=self.business_id,
                app_id=app_id,
            )
            report_ids = report_ids + [
                app_report['id']
                for app_report in app_reports
            ]
        return report_ids

    # TODO paginate
    def get_business_id_by_report(self, report_id: str, **kwargs) -> str:
        """Bottom-up method
        Having a report_id return the app it belongs to
        """
        app_id: str = self.get_app_id_by_report(report_id=report_id, **kwargs)
        business_id: str = self.get_business_id_by_app(app_id=app_id, **kwargs)
        return business_id


class CascadeCreateExplorerAPI(CreateExplorerAPI):
    update_report = UpdateExplorerAPI.update_report

    def __init__(self, api_client):
        self.api_client = api_client

    def create_app_from_app_type_normalized_name(self, app_type_name: str) -> Dict:
        """Create AppType and App if required and return the App component
        """
        try:
            app_type: Dict = self._create_app_type(name=app_type_name)
        except ValueError:  # It already exists then
            app_type: Dict = (
                self._find_app_type_by_name_filter(name=app_type_name)
            )

        app_type_id: str = app_type['id']
        apps: Dict = self._get_business_apps(business_id=self.business_id)
        target_apps = [app for app in apps if app['appType']['id'] == app_type_id]

        if not apps:
            app: Dict = (
                self._create_app(
                    business_id=self.business_id,
                    app_type_id=app_type_id,
                )
            )
        else:
            app: Dict = target_apps[0]
        return app

    def create_report_and_dataset(
        self, business_id: str, app_id: str,
        report_metadata: Dict,
        items: Union[List[str], Dict],
        report_properties: Dict,
        report_dataset_properties: Optional[Dict] = None,
        sort: Optional[Dict] = None,
        real_time: bool = False,
    ) -> Dict[str, Union[Dict, List[Dict]]]:
        """
        Example
        -------
        sort = {
            'field': 'date'
            'direction': 'asc',
        }

        1. Create a report
        2. Create a dataset
        3. Create data associated to a dataset
        4. Associate dataset and report through reportDataSet
        """
        report: Dict = self.create_report(
            business_id=business_id,
            app_id=app_id,
            report_metadata=report_metadata,
            real_time=real_time,
        )

        dataset: Dict = self.create_dataset(business_id=business_id)
        dataset_id: str = dataset['id']

        if type(items) == list:  # ECHARTS2
            items_keys: Optional[List[str]] = list(items[0].keys())
            report_dataset_properties = {'mapping': items_keys}
            if sort:
                report_dataset_properties.update(sort)
        elif type(items) == dict:  # FORM
            items_keys: Optional[List[str]] = None
            items = [items]
            try:
                assert report_dataset_properties is not None
            except AssertionError:
                raise ValueError(
                    'report_dataset_properties is required if items is a dict'
                )
        else:
            raise ValueError('items must be a list or dict')

        report_dataset: Dict = self.create_reportdataset(
            business_id=business_id,
            app_id=app_id,
            report_id=report['id'],
            dataset_id=dataset_id,
            dataset_properties=json.dumps(report_dataset_properties),
        )

        data: List[Dict] = self.create_data_points(
            business_id=business_id,
            dataset_id=dataset_id,
            items=items,
        )

        # Syntax to be accepted by the FrontEnd
        options_dataset_id: str = '#{' + f'{report_dataset["id"]}' + '}'
        report_properties['dataset'] = {'source': options_dataset_id}
        if items_keys is not None:  # ECHARTS2
            report_properties['dimensions']: items_keys
        report_properties = {'properties': json.dumps({'option': report_properties})}

        report: Dict = self.update_report(
            business_id=business_id,
            app_id=app_id,
            report_id=report['id'],
            report_metadata=report_properties,
        )

        return {
            'report': report,
            'dataset': dataset,
            'report_dataset': report_dataset,
            'data': data,
        }


class DeleteExplorerApi(MultiCascadeExplorerAPI, UpdateExplorerAPI):
    """Get Businesses, Apps, Paths and Reports in any possible combination
    """

    def __init__(self, api_client):
        super().__init__(api_client)

    def delete_business(self, business_id: str):
        """Delete a Business.
        All apps, reports and data associated with that business is removed by the API
        """
        endpoint: str = f'business/{business_id}'
        self.api_client.query_element(
            method='DELETE', endpoint=endpoint,
        )

    def delete_app_type(self, app_type_id: str):
        """Delete an appType"""
        endpoint: str = f'apptype/{app_type_id}'
        self.api_client.query_element(
            method='DELETE', endpoint=endpoint,
        )

    def delete_app(self, business_id: str, app_id: str) -> Dict:
        """Delete an App
        All reports and data associated with that app is removed by the API
        """
        endpoint: str = f'business/{business_id}/app/{app_id}'
        result: Dict = self.api_client.query_element(
            method='DELETE', endpoint=endpoint
        )
        return result

    def delete_path(self, business_id: str, app_id: str, path_name: str):
        """Delete all Reports in a path
        All data associated with that report is removed by the API"""
        reports: List[Dict] = (
            self.get_path_reports(
                business_id=business_id,
                app_id=app_id,
                path_name=path_name,
            )
        )
        for report in reports:
            report_id: str = report['id']
            self.delete_report_and_entries(report_id)

    def delete_report(
        self, business_id: str, app_id: str, report_id: str,
        relocating: bool = True, delete_data: bool = True,
    ) -> Dict:
        """Delete a Report, relocating reports underneath to avoid errors
        """
        reports: List[Dict] = (
            self._get_app_reports(
                business_id=business_id,
                app_id=app_id
            )
        )
        target_report: Dict = self.get_report(
            business_id=business_id,
            app_id=app_id,
            report_id=report_id,
        )
        target_report_grid: str = target_report.get('grid')

        # TO BE deprecated with row, column and grid!
        # TODO this looks like a different method
        if target_report_grid:
            target_report_row: int = int(target_report_grid.split(',')[0])
            for report in reports:
                report_grid: str = report.get('grid')
                if report_grid:
                    report_row: int = int(report_grid.split(',')[0])
                    if report_row > target_report_row:
                        report_row -= 1
                        report_column: int = int(report.get('grid').split(',')[1])
                        grid: str = f'{report_row}, {report_column}'
                        self.update_report(
                            business_id=business_id,
                            app_id=app_id, report_id=report_id,
                            report_metadata={'grid': grid},
                        )

        endpoint: str = f'business/{business_id}/app/{app_id}/report/{report_id}'
        result: Dict = self.api_client.query_element(
            method='DELETE', endpoint=endpoint
        )
        return result

    def delete_reportdataset(
            self, business_id: str, app_id: str,
            report_id: str, reportdataset_id: str,
    ) -> Dict:
        """"""
        endpoint: str = (
            f'business/{business_id}/'
            f'app/{app_id}/'
            f'report/{report_id}/'
            f'{reportdataset_id}'
        )
        result: Dict = self.api_client.query_element(
            method='DELETE', endpoint=endpoint
        )
        return result

    def delete_dataset(self, business_id: str, dataset_id: str) -> Dict:
        """"""
        endpoint: str = f'business/{business_id}/dataset/{dataset_id}'
        result: Dict = self.api_client.query_element(
            method='DELETE', endpoint=endpoint
        )
        return result

    def delete_report_entries(
        self, business_id: str, app_id: str, report_id: str,
    ) -> None:
        """Delete a Report, relocating reports underneath to avoid errors
        """
        report_entries: List[Dict] = (
            self.get_report_data(
                business_id=business_id,
                app_id=app_id,
                report_id=report_id,
            )
        )

        for report_entry in report_entries:
            report_entry_id: str = report_entry['id']
            endpoint: str = (
                f'business/{business_id}/'
                f'app/{app_id}/'
                f'report/{report_id}/'
                f'reportEntry/{report_entry_id}'
            )
            _: Dict = self.api_client.query_element(
                method='DELETE', endpoint=endpoint
            )

    def delete_file(
        self, business_id: str, app_id: str, file_id: str,
    ) -> Dict:
        """Delete a file
        """
        endpoint: str = f'business/{business_id}/app/{app_id}/file/{file_id}'
        result: Dict = self.api_client.query_element(method='DELETE', endpoint=endpoint)
        return result


# TODO los siguientes puntos:
#  . Si elimino (delete) un report se eliminan sus reportdataset?
#  . Tengo función cascade para dado un report coger (GET) todos los reportdataset?
#  . No puedo crear data sin un dataset asociado, esto es así?
#  . Si elimino (delete) un dataset se eliminan sus data ?
#  . Tengo función cascade para dado un dataset coger (GET) todos los data?
class MultiDeleteApi:
    """Get Businesses, Apps, Paths and Reports in any possible combination
    """
    _get_business = GetExplorerAPI.get_business
    _get_app_type = GetExplorerAPI.get_app_type
    _get_app = GetExplorerAPI.get_app

    _delete_business = DeleteExplorerApi.delete_business
    _delete_app = DeleteExplorerApi.delete_app
    _delete_app_type = DeleteExplorerApi.delete_app_type
    _delete_report = DeleteExplorerApi.delete_report
    _delete_dataset = DeleteExplorerApi.delete_dataset

    def __init__(self):
        return

    def _delete_business_and_app_type(
        self, business_id: str, app_type_id: str
    ):
        try:
            self._delete_business(business_id)
        except Exception as e_bd:
            raise ValueError(
                f'{e_bd} | Nor Business nor AppType were deleted | ' 
                f'business_id: {business_id} | '
                f'app_type_id: {app_type_id}'
            )

        try:
            _ = self._get_business(business_id)
        except ApiClientError:
            pass
        except Exception as e_gb:
            raise ValueError(
                f'{e_gb} | Nor Business nor AppType were deleted | '
                f'business_id: {business_id} | app_type_id: {app_type_id}'
            )

        try:
            self._delete_app_type(app_type_id)
        except ApiClientError:
            return {}
        except Exception as e_atd:
            raise ValueError(
                f'{e_atd} | AppType was not deleted | '
                f'app_type_id: {app_type_id}'
            )

        try:
            _ = self._get_app_type(app_type_id)
        except ApiClientError:
            return {}
        except Exception as e_atg:
            raise ValueError(
                f'{e_atg} | AppType was not deleted | '
                f'app_type_id: {app_type_id}'
            )

    def _delete_business_and_app(
        self, business_id: str, app_id: str,
    ):
        try:
            self._delete_business(business_id)
        except Exception as e_bd:
            raise ValueError(
                f'{e_bd} | Nor Business nor App were deleted | ' 
                f'business_id: {business_id} | '
                f'app_id: {app_id}'
            )

        try:
            _ = self._get_business(business_id)
        except ApiClientError:
            pass
        except Exception as e_gb:
            raise ValueError(
                f'{e_gb} | Nor Business nor App were deleted | '
                f'business_id: {business_id} | '
                f'app_id: {app_id}'
            )

        try:
            self._delete_app(app_id)
        except ApiClientError:
            return {}
        except Exception as e_atd:
            raise ValueError(
                f'{e_atd} | App was not deleted | '
                f'app_id: {app_id}'
            )

        try:
            _ = self._get_app(app_id)
        except ApiClientError:
            return {}
        except Exception as e_atg:
            raise ValueError(
                f'{e_atg} | App was not deleted | '
                f'app_id: {app_id}'
            )

    def delete_report_and_dataset(
            self, business_id: str, app_id: str, report_id: str, dataset_id: str,
    ):
        self._delete_report(
            business_id=business_id,
            app_id=app_id,
            report_id=report_id
        )
        self._delete_dataset(business_id=business_id, dataset_id=dataset_id)


class MultiCreateApi(MultiDeleteApi):
    """If some upper level elements are not created it does it
    """
    _get_universe_app_types = CascadeExplorerAPI.get_universe_app_types
    _get_app_by_type = CascadeExplorerAPI.get_app_by_type

    _create_business = CascadeCreateExplorerAPI.create_business
    _create_app_type = CascadeCreateExplorerAPI.create_app_type
    _create_app = CascadeCreateExplorerAPI.create_app
    _create_report = CascadeCreateExplorerAPI.create_report

    def __init__(self):
        super().__init__()

    def create_business_and_app(
        self, app_type_id: str, business_name: str, app_metadata: Dict,
    ) -> Dict[str, Dict]:
        """Create new Report associated to an AppId

        :param app_type_id:
        :param business_name:
        :param app_metadata:
        """
        business: Dict = self._create_business(name=business_name)
        business_id: str = business['id']

        try:
            app: Dict = (
                self._create_app(
                    business_id=business_id,
                    app_type_id=app_type_id,
                    app_metadata=app_metadata,
                )
            )
        except Exception as e:
            self._delete_business(business_id=business_id)
            try:
                _ = self._get_business(business_id)
                raise ValueError(
                    f'{e} | The app was not created but a new business did '
                    f'that probably should be deleted manually with id '
                    f'{business_id}'
                )
            except ApiClientError:
                return {}

        return {
            'business': business,
            'app': app,
        }

    def create_app_type_and_app(
        self, business_id: str,
        app_type_metadata: Dict,
        app_metadata: Optional[Dict] = None,
    ) -> Dict[str, Dict]:
        """
        If app_type_id is None we create it
        """
        try:
            app_type: Dict = self._create_app_type(**app_type_metadata)
        except ValueError:
            app_type_name: str = app_type_metadata['name']
            app_type: Dict = self._get_app_type_by_name(app_type_name)

        app_type_id: str = app_type['id']
        app_metadata['app_type_id'] = app_type_id
        app_metadata['business_id'] = business_id

        app: Dict = self._get_app_by_type(
            business_id=business_id,
            app_type_id=app_type_id,
        )
        if not app:
            if app_metadata.get('name'):
                app: Dict = self._create_app(**app_metadata)
            else:  # get the AppType name and use it
                app_metadata.update({'name': app_type_metadata['name']})
                app: Dict = self._create_app(**app_metadata)

        return {
            'app_type': app_type,
            'app': app
        }

    def create_app_and_report(
        self, business_id: str, app_type_id: str,
        app_metadata: Dict, report_metadata: Dict,
    ) -> Dict:
        """Create new Report associated to an AppId

        :param business_id:
        :param app_type_id:
        :param app_metadata:
        :param report_metadata: A dict with all the values required to create a report
        """
        app: Dict = (
            self._create_app(
                business_id=business_id,
                app_type_id=app_type_id,
                app_metadata=app_metadata,
            )
        )
        app_id: str = app['id']

        try:
            report: Dict = (
                self._create_report(
                    business_id=business_id,
                    app_id=app_id,
                    report_metadata=report_metadata,
                )
            )
        except Exception as e:
            raise f'{e} | app_id created: {app_id} | Better delete it'

        return report

    def create_business_app_and_app_type(
        self, business_name: str,
        app_metadata: Dict,
        app_type_metadata: Dict,
    ) -> Dict[str, Dict]:
        """
        """
        app_type: Dict = self._create_app_type(**app_type_metadata)
        app_type_id: str = app_type['id']
        app_metadata['app_type_id'] = app_type_id

        business: Dict = {}
        try:
            business: Dict = self._create_business(business_name)
            business_id: str = business['id']
            app_metadata['business_id'] = business_id
        except Exception as e:
            try:
                self._delete_app_type(app_type_id=app_type_id)
            except ApiClientError:
                return {}
            except Exception as e:
                raise ValueError(
                    f'Business was not created | '
                    f'AppType was created with app_type_id = {app_type_id}'
                    f'App was not created | '
                )

        app: Dict = {}
        try:
            app: Dict = self._create_app(**app_metadata)
        except Exception as e:
            try:
                self._delete_business_and_app_type(
                    business_id=self.business_id,
                    app_type_id=app_type_id,
                )
            except ApiClientError:
                return {}
            except Exception as e:
                raise ValueError(f'App was not created | {e}')

        return {
            'business': business,
            'app_type': app_type,
            'app': app
        }

    def create_business_app_and_report(
        self, app_type_id: str,
        business_name: str,
        app_metadata: Dict,
        report_metadata: Dict,
    ) -> Dict[str, Dict]:
        """
        """
        business: Dict = self.create_business(business_name)
        business_id: str = business['id']
        app_metadata['business_id'] = business_id
        app_metadata['app_type_id'] = app_type_id

        try:
            app: Dict = self.create_app(
                business_id=business_id,
                app_metadata=app_metadata,
            )
            app_id = app['id']
        except Exception as e:
            try:
                self.delete_business(business_id)
            except ApiClientError:
                return {}
            except Exception as e:
                raise ValueError(
                    f'{e} | Business with business_id {business_id} '
                    f'created that probably wants to be removed | '
                    f'App was not created | '
                    f'Report was not created'
                )

        try:
            report: Dict = self.create_report(
                business_id=business_id,
                app_id=app_id,
                report_metadata=report_metadata,
            )
        except Exception as e:
            try:
                self._delete_business_and_app(
                    business_id=business_id,
                    app_id=app_id,
                )
            except ApiClientError:
                return {}
            except Exception as e_dba:
                raise ValueError(
                    f'{e} | {e_dba} | Report was not created'
                )
            return {}

        return {
            'business': business,
            'app': app,
            'report': report,
        }

    def create_business_app_type_app_and_report(
        self, business_name: str,
        app_type_metadata: Dict,
        app_metadata: Dict,
        report_metadata: Dict,
    ) -> Dict[str, Dict]:
        """
        """
        d = self.create_business_app_and_app_type(
            business_name=business_name,
            app_type_metadata=app_type_metadata,
            app_metadata=app_metadata,
        )
        business_id: str = d['business']['id']
        app_id: str = d['app']['id']

        try:
            report: Dict = self.create_report(
                business_id=business_id,
                app_id=app_id,
                report_metadata=report_metadata,
            )
        except Exception as e:
            try:
                self._delete_business_and_app(
                    business_id=business_id,
                    app_id=app_id,
                )
            except ApiClientError:
                return {}
            except Exception as e_:
                raise ValueError(
                    f'{e} | {e_} | Report was not created'
                )

            app_type_id: str = d['app_type']['id']
            try:
                self.delete_app_type(app_type_id)
            except Exception as e_:
                raise ValueError(
                    f'{e_} | Report was not created | '
                    f'App type was created with app_type_id: {app_type_id}'
                )
            return {}

        return {
            'app_type': d['app_type'],
            'business': d['business'],
            'app': d['app'],
            'report': report,
        }


class UniverseExplorerApi:
    """"""
    get_universe_businesses = CascadeExplorerAPI.get_universe_businesses
    get_universe_app_types = CascadeExplorerAPI.get_universe_app_types


class BusinessExplorerApi:
    """"""
    get_business = GetExplorerAPI.get_business
    get_universe_businesses = CascadeExplorerAPI.get_universe_businesses
    _find_business_by_name_filter = CascadeExplorerAPI.find_business_by_name_filter
    create_business = CascadeCreateExplorerAPI.create_business
    update_business = UpdateExplorerAPI.update_business

    get_business_apps = CascadeExplorerAPI.get_business_apps
    get_business_app_ids = CascadeExplorerAPI.get_business_app_ids
    get_business_all_apps_with_filter = CascadeExplorerAPI.get_business_apps_with_filter

    delete_business = DeleteExplorerApi.delete_business


class AppTypeExplorerApi:
    """"""
    _create_normalized_name = CreateExplorerAPI._create_normalized_name
    _create_key_name = CreateExplorerAPI._create_key_name

    get_app_type = GetExplorerAPI.get_app_type
    get_universe_app_types = CascadeExplorerAPI.get_universe_app_types
    _find_app_type_by_name_filter = CascadeExplorerAPI.find_app_type_by_name_filter
    create_app_type = CascadeCreateExplorerAPI.create_app_type
    update_app_type = UpdateExplorerAPI.update_app_type

    delete_app_type = DeleteExplorerApi.delete_app_type


class AppExplorerApi:
    _create_normalized_name = CreateExplorerAPI._create_normalized_name
    _create_key_name = CreateExplorerAPI._create_key_name

    get_app = GetExplorerAPI.get_app
    create_app = CascadeCreateExplorerAPI.create_app
    update_app = UpdateExplorerAPI.update_app

    _get_business_apps = CascadeExplorerAPI.get_business_apps
    get_business_apps = CascadeExplorerAPI.get_business_apps
    find_app_by_name_filter = CascadeExplorerAPI.find_app_by_name_filter
    get_app_reports = CascadeExplorerAPI.get_app_reports
    get_app_report_ids = CascadeExplorerAPI.get_app_report_ids
    get_app_path_names = CascadeExplorerAPI.get_app_path_names
    get_app_reports_by_filter = MultiCascadeExplorerAPI.get_app_reports_by_filter
    get_app_by_type = CascadeExplorerAPI.get_app_by_type
    get_app_type = CascadeExplorerAPI.get_app_type
    get_app_by_name = CascadeExplorerAPI.get_app_by_name
    get_app_by_url = CascadeExplorerAPI.get_app_by_url

    delete_app = DeleteExplorerApi.delete_app


class ReportExplorerApi:

    get_report = GetExplorerAPI.get_report
    get_report_data = GetExplorerAPI.get_report_data
    _get_report_with_data = GetExplorerAPI._get_report_with_data

    _get_app_reports = CascadeExplorerAPI.get_app_reports

    create_report = CascadeCreateExplorerAPI.create_report
    create_app_and_report = MultiCreateApi.create_app_and_report

    update_report = UpdateExplorerAPI.update_report

    get_business_id_by_report = MultiCascadeExplorerAPI.get_business_id_by_report

    delete_report = DeleteExplorerApi.delete_report


class DatasetExplorerApi:

    get_dataset = GetExplorerAPI.get_dataset

    get_dataset_data = CascadeExplorerAPI.get_dataset_data

    create_data_points = CreateExplorerAPI.create_data_points
    create_dataset = CascadeCreateExplorerAPI.create_dataset

    update_dataset = UpdateExplorerAPI.update_dataset

    delete_dataset = DeleteExplorerApi.delete_dataset


class ReportDatasetExplorerApi:

    get_reportdataset = GetExplorerAPI.get_reportdataset
    get_report_datasets = CascadeExplorerAPI.get_report_datasets
    get_report_dataset_data = CascadeExplorerAPI.get_report_dataset_data

    create_reportdataset = CascadeCreateExplorerAPI.create_reportdataset
    create_report_and_dataset = CascadeCreateExplorerAPI.create_report_and_dataset

    update_reportdataset = UpdateExplorerAPI.update_reportdataset

    delete_reportdataset = DeleteExplorerApi.delete_reportdataset
    delete_report_and_dataset = MultiDeleteApi.delete_report_and_dataset


class FileExplorerApi:
    _get_file = GetExplorerAPI.get_file
    get_files = GetExplorerAPI.get_files

    _create_file = CreateExplorerAPI.create_file

    _delete_file = DeleteExplorerApi.delete_file

    _get_business_apps = CascadeExplorerAPI.get_business_apps
    _get_app_by_name = CascadeExplorerAPI.get_app_by_name
    get_business_apps = CascadeExplorerAPI.get_business_apps


class ExplorerApi(
    CascadeCreateExplorerAPI,
    DeleteExplorerApi,
):
    """Get Businesses, Apps, Paths and Reports in any possible combination
    """

    def __init__(self, api_client):
        super().__init__(api_client)

    # TODO WiP
    def has_app_report_data(self, business_id: str, app_id: str) -> bool:
        """"""
        report_ids: List[str] = self.get_app_report_ids(
            business_id=business_id, app_id=app_id
        )
        for report_id in report_ids:
            result: bool = self.has_report_report_entries(report_id)
            if result:
                return True
        return False

    # TODO WiP
    def has_path_data(self, business_id: str, app_id: str, path_name: str) -> bool:
        """"""
        report_ids: List[str] = self.get_app_report_ids(
            business_id=business_id, app_id=app_id
        )
        for report_id in report_ids:
            result: bool = self.has_report_report_entries(report_id)
            if result:
                return True
        return False



from os import getenv
import datetime as dt
import shimoku_api_python as shimoku

#  Environment variables (credentials):

api_key: str = getenv('API_TOKEN')
universe_id: str = getenv('UNIVERSE_ID')
business_id: str = getenv('BUSINESS_ID')
environment: str = getenv('ENVIRONMENT')

config = {
    'access_token': api_key,
}

s = shimoku.Client(
    config=config,
    universe_id=universe_id,
    environment=environment,
)

s.plt.set_business(business_id=business_id)

#  disappear = s.plt.delete_path(menu_path='testRM')

menu_path: str = 'Test 1'

data_ = [
    {
        'description': 'cryo',
        'title': 'VAC',
        'value': '15',
        'color': 'success-background'
    }
]

s.plt.indicator(
    data=data_,
    menu_path=menu_path,
    row=1, column=1,
    value='value',
    color='color',
    header='title',
    footer='description',
)

data_ = [
    {
        'description': 'multi',
        'title': 'VAC',
        'value': 'distributed',
        'color': 'error-background',
    }
]
s.plt.indicator(
    data=data_,
    menu_path=menu_path,
    row=2, column=1,
    value='value',
    color='color',
    header='title',
    footer='description',
)


data = [
    {'date': dt.date(2021, 1, 1), 'x': 5, 'y': 5},
    {'date': dt.date(2021, 1, 2), 'x': 6, 'y': 5},
    {'date': dt.date(2021, 1, 3), 'x': 4, 'y': 5},
    {'date': dt.date(2021, 1, 4), 'x': 7, 'y': 5},
    {'date': dt.date(2021, 1, 5), 'x': 3, 'y': 5},
]

s.plt.bar(
        data=data,
        x='date', y=['x', 'y'],
        menu_path=menu_path,
        row=3, column=1,
    )



""""""

from abc import ABC

from shimoku_api_python.api.explorer_api import UniverseExplorerApi


class UniverseMetadataApi(UniverseExplorerApi, ABC):
    """
    """
    def __init__(self, api_client):
        self.api_client = api_client



""""""
from os import getenv

import shimoku_api_python as shimoku

api_key: str = getenv('API_TOKEN')
base_url: str = getenv('BASE_URL')
universe_id: str = getenv('UNIVERSE_ID')
business_id: str = getenv('BUSINESS_ID')
app_id: str = getenv('APP_ID')


config = {
    'access_token': api_key,
}

s = shimoku.Client(
    config=config,
    universe_id=universe_id,
)


def test_request():
    result = s.api_client.request(
        method='GET',
        url=f'{base_url}{business_id}/app/{app_id}/',
    )
    assert result.status_code == 200
    print(result)


def test_query_element():
    result = s.api_client.query_element(
        method='GET',
        element_name=f'{business_id}/app',
        element_id=app_id,
    )
    assert result.status_code == 200
    print(result)


def test_ping():
    is_alive: bool = s.ping()
    assert is_alive


test_request()
test_query_element()
test_ping()



"""
Used as base Mailchimp:
https://github.com/mailchimp/mailchimp-marketing-python/blob/master/mailchimp_marketing/api_client.py
"""


from typing import List, Dict

import datetime
import requests
import json

from shimoku_api_python.exceptions import ApiClientError


class ApiClient(object):
    PRIMITIVE_TYPES = (float, int, bool, bytes, str)

    def __init__(self, universe_id: str, environment: str, config={}):
        if environment == 'production':
            self.host = 'https://api.shimoku.io/external/v1/'
        elif environment == 'staging':
            self.host = 'https://api.staging.shimoku.io/external/v1/'
        elif environment == 'develop':
            self.host = 'https://api.develop.shimoku.io/external/v1/'
        else:
            raise ValueError(
                f'The namespace must be either "production", "staging" or "develop | '
                f'namespace introduced: {environment}'
            )

        self.host: str = f'{self.host}universe/{universe_id}/'

        # DEFAULTS
        # Api key
        self.api_key: str = ''
        self.is_basic_auth: bool = False
        # OAuth
        self.access_token: str = ''
        self.is_oauth: bool = False
        # General
        self.server: str = 'invalid-server'
        self.timeout: int = 120

        self.default_headers = {
            'Content-Type': 'application/json',
            'User-Agent': 'Swagger-Codegen/0.0/python'
        }
        self.set_config(config)

        # Default vars

    def set_config(self, config={}):
        """Set all config values"""
        # Basic Auth
        self.api_key: str = config['api_key'] if 'api_key' in config.keys() else ''
        self.is_basic_auth: bool = self.api_key != ''

        # OAuth
        self.access_token: str = config['access_token'] if 'access_token' in config.keys() else ''
        self.is_oauth: bool = self.access_token != ''

        if not self.is_oauth and not self.is_basic_auth:
            raise ValueError('You must provide either an API Key or Access Token')

        # If using Basic auth and no server is provided,
        # attempt to extract it from the api_key directly.
        self.server: str = config['server'] if 'server' in config.keys() else 'invalid-server'
        if self.server == 'invalid-server' and self.is_basic_auth:
            self.server: str = self.get_server_from_api_key(self.api_key)

        self.timeout = config['timeout'] if 'timeout' in config.keys() else 120

    def call_api(
            self, resource_path, method, path_params=None, query_params=None,
            header_params=None, body=None, collection_formats=None, **kwargs
    ):
        """Create and call the API request with headers, params and others"""
        # header parameters
        header_params = header_params or {}
        header_params.update(self.default_headers)
        if header_params:
            header_params = self.sanitize_for_serialization(header_params)
            header_params = dict(self.parameters_to_tuples(header_params,
                                                           collection_formats))

        # path parameters
        if path_params:
            path_params = self.sanitize_for_serialization(path_params)
            path_params = self.parameters_to_tuples(path_params, collection_formats)
            for k, v in path_params:
                # specified safe chars, encode everything
                resource_path = resource_path.replace(
                    '{%s}' % k,
                    str(v)
                )

        # query parameters
        if query_params:
            query_params = self.sanitize_for_serialization(query_params)
            query_params = self.parameters_to_tuples(query_params, collection_formats)

        # request url
        url = self.host + resource_path

        if self.server:
            url = url.replace('server', self.server)

        # perform request and return response
        try:
            res = self.request(
                method, url, query_params,
                headers=header_params, body=body,
            )
        except Exception as err:
            raise ApiClientError(err)

        try:
            if 'application/json' in res.headers.get('content-type'):
                data = res.json()
            else:
                data = res.text
        except Exception:
            data = None

        if res.ok:
            return data
        else:
            raise ApiClientError(data)

    def set_http_info(self, **kwargs):  # noqa: E501
        """
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.remove_with_http_info(campaign_id, async_req=True)
        >>> result = thread.get()
        :param async_req bool
        :param str campaign_id: The unique id for the campaign. (required)
        :return: None
                 If the method is called asynchronously,
                 returns the request thread.
        """
        params = locals()
        for key, val in params['kwargs'].items():
            params[key] = val
        body_params = params['kwargs'].get('data')
        del params['kwargs']

        collection_formats = {}

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        # HTTP header `Accept`
        header_params['Accept'] = self.select_header_accept(
            ['application/json', 'application/problem+json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = ['basicAuth']  # noqa: E501

        return (
            query_params, header_params,
            kwargs.get('body_params'), form_params, local_var_files,
            auth_settings, params, collection_formats,
        )

    def query_element(
        self, method: str, endpoint: str, **kwargs
    ) -> Dict:
        """Retrieve an element if the endpoint exists

        :param method: examples are 'GET', 'POST', etc
        :param endpoint: example: 'business/{businessId}/app/{appId}
        """
        (
            query_params, header_params,
            body_params, form_params, local_var_files,
            auth_settings, params, collection_formats,
        ) = self.set_http_info(**kwargs)

        path_params = {}
        if endpoint in params:
            path_params[endpoint] = params[endpoint]  # noqa: E501

        element_data: Dict = (
            self.call_api(
                endpoint, method,
                path_params,
                query_params,
                header_params,
                body=body_params,
                post_params=form_params,
                files=local_var_files,
                response_type=None,  # noqa: E501
                auth_settings=auth_settings,
                async_req=params.get('async_req'),
                _return_http_data_only=params.get('_return_http_data_only'),
                _preload_content=params.get('_preload_content', True),
                _request_timeout=params.get('_request_timeout'),
                collection_formats=collection_formats
            )
        )
        return element_data

    def request(self, method, url, query_params=None, headers=None, body=None):
        auth = None

        if self.is_basic_auth:
            auth = ('user', self.api_key)

        if self.is_oauth:
            if headers:
                headers.update({'Authorization': 'Bearer ' + self.access_token})
            else:
                headers = {'Authorization': 'Bearer ' + self.access_token}

        if method == "GET":
            return requests.get(
                url, params=query_params, headers=headers,
                auth=auth, timeout=self.timeout
            )
        elif method == "HEAD":
            return requests.head(
                url, params=query_params, headers=headers,
                auth=auth, timeout=self.timeout
            )
        elif method == "OPTIONS":
            return requests.options(
                url, params=query_params, headers=headers,
                auth=auth, timeout=self.timeout
            )
        elif method == "POST":
            return requests.post(
                url, data=json.dumps(body), params=query_params,
                headers=headers, auth=auth, timeout=self.timeout
            )
        elif method == "PUT":
            return requests.put(
                url, data=json.dumps(body), params=query_params,
                headers=headers, auth=auth, timeout=self.timeout
            )
        elif method == "PATCH":
            return requests.patch(
                url, data=json.dumps(body), params=query_params,
                headers=headers, auth=auth, timeout=self.timeout
            )
        elif method == "DELETE":
            return requests.delete(
                url, params=query_params, headers=headers,
                auth=auth, timeout=self.timeout
            )
        else:
            raise ValueError(
                "http method must be `GET`, `HEAD`, `OPTIONS`,"
                " `POST`, `PATCH`, `PUT` or `DELETE`."
            )

    def raw_request(self, **kwargs):
        return requests.request(**kwargs)

    def sanitize_for_serialization(self, obj):
        """Builds a JSON POST object.
        If obj is None, return None.
        If obj is str, int, long, float, bool, return directly.
        If obj is datetime.datetime, datetime.date
            convert to string in iso8601 format.
        If obj is list, sanitize each element in the list.
        If obj is dict, return the dict.
        If obj is swagger model, return the properties dict.
        :param obj: The data to serialize.
        :return: The serialized form of data.
        """
        if obj is None:
            return None
        elif isinstance(obj, self.PRIMITIVE_TYPES):
            return obj
        elif isinstance(obj, list):
            return [self.sanitize_for_serialization(sub_obj)
                    for sub_obj in obj]
        elif isinstance(obj, tuple):
            return tuple(self.sanitize_for_serialization(sub_obj)
                         for sub_obj in obj)
        elif isinstance(obj, (datetime.datetime, datetime.date)):
            return obj.isoformat()

        if isinstance(obj, dict):
            obj_dict = obj
        else:
            # Convert model obj to dict except
            # attributes `swagger_types`, `attribute_map`
            # and attributes which value is not None.
            # Convert attribute name to json key in
            # model definition for request.
            obj_dict = {
                obj.attribute_map[attr]: getattr(obj, attr)
                for attr, _ in obj.items()
                if getattr(obj, attr) is not None
            }

        return {key: self.sanitize_for_serialization(val)
                for key, val in obj_dict.items()}

    @staticmethod
    def parameters_to_tuples(params, collection_formats):
        """Get parameters as list of tuples, formatting collections.
        :param params: Parameters as dict or list of two-tuples
        :param dict collection_formats: Parameter collection formats
        :return: Parameters as list of tuples, collections formatted
        """
        new_params = []
        if collection_formats is None:
            collection_formats = {}
        for k, v in params.items() if isinstance(params, dict) else params:  # noqa: E501
            if k in collection_formats:
                collection_format = collection_formats[k]
                if collection_format == 'multi':
                    new_params.extend((k, value) for value in v)
                else:
                    if collection_format == 'ssv':
                        delimiter = ' '
                    elif collection_format == 'tsv':
                        delimiter = '\t'
                    elif collection_format == 'pipes':
                        delimiter = '|'
                    else:  # csv is the default
                        delimiter = ','
                    new_params.append(
                        (k, delimiter.join(str(value) for value in v)))
            else:
                new_params.append((k, v))
        return new_params

    @staticmethod
    def select_header_accept(accepts):
        """Returns `Accept` based on an array of accepts provided.
        :param accepts: List of headers.
        :return: Accept (e.g. application/json).
        """
        if not accepts:
            return

        accepts = [x.lower() for x in accepts]

        if 'application/json' in accepts:
            return 'application/json'
        else:
            return ', '.join(accepts)

    @staticmethod
    def get_server_from_api_key(api_key: str) -> str:
        try:
            split: List[str] = api_key.split('-')
            if len(split) == 2:
                return split[1]
            else:
                return 'invalid-server'
        except:
            return ''

    @staticmethod
    def select_header_content_type(content_types):
        """Returns `Content-Type` based on an array of content_types provided.
        :param content_types: List of content-types.
        :return: Content-Type (e.g. application/json).
        """
        if not content_types:
            return 'application/json'

        content_types = [x.lower() for x in content_types]

        if 'application/json' in content_types or '*/*' in content_types:
            return 'application/json'
        else:
            return content_types[0]



""""""
from os import getenv

import shimoku_api_python as shimoku

api_key: str = getenv('API_TOKEN')
base_url: str = getenv('BASE_URL')
universe_id: str = getenv('UNIVERSE_ID')
business_id: str = getenv('BUSINESS_ID')
app_id: str = getenv('APP_ID')


config = {
    'access_token': api_key,
}

s = shimoku.Client(
    config=config,
    universe_id=universe_id,
)


def test_create_business_and_app():
    # TODO it must create and delete
    # s.wildcards.create_business_and_app
    raise NotImplementedError


def test_create_app_type_and_app():
    # TODO it must create and delete
    # s.wildcards.create_app_type_and_app
    raise NotImplementedError


def test_create_app_and_report():
    # TODO it must create and delete
    # s.wildcards.create_app_and_report
    raise NotImplementedError


def test_create_business_app_and_report():
    # TODO it must create and delete
    # s.wildcards.create_business_app_and_report
    raise NotImplementedError


def test_create_business_app_type_app_and_report():
    # TODO it must create and delete
    # s.wildcards.create_business_app_type_app_and_report
    raise NotImplementedError


test_create_business_and_app()
test_create_app_type_and_app()
test_create_app_and_report()
test_create_business_app_and_report()
test_create_business_app_type_app_and_report()



""""""

from abc import ABC
from typing import List, Dict, Optional, Union, Any, Callable
import re
from io import StringIO
import pickle

import datetime as dt

import pandas as pd
from shimoku_api_python.api.explorer_api import FileExplorerApi


class BasicFileMetadataApi(FileExplorerApi, ABC):
    """
    """
    def __init__(self, api_client):
        self.api_client = api_client

    @staticmethod
    def _encode_file_name(
            file_name: str, date: dt.date, chunk: Optional[int] = None
    ) -> str:
        """Append the date and chunk to the file name"""
        return (
            f'{file_name}_date:{date.isoformat()}_chunk:{chunk}'
            if chunk is not None else f'{file_name}_date:{date.isoformat()}'
        )

    @staticmethod
    def _decode_file_name(file_name: str) -> str:
        """Decode the file name into its components
        """
        file_name_date_parts: List[str] = file_name.split('_date:')
        file_name_chunk_parts: List[str] = file_name.split('_chunk:')

        if len(file_name_date_parts) == 1 and len(file_name_chunk_parts) == 1:
            return file_name
        elif len(file_name_date_parts) == 2 or len(file_name_chunk_parts) == 1:
            return file_name_date_parts[0]
        else:
            raise ValueError('Invalid file name has multiple times "_date:" or "_chunk:"')

    @staticmethod
    def _get_file_date(file_name: str) -> Optional[dt.date]:
        try:
            date_iso: str = file_name.split('_date:')[1].split('_chunk:')[0]
        except IndexError:
            return None
        return dt.datetime.strptime(date_iso, '%Y-%m-%d').date()

    # TODO applies only to dataframes
    def _get_file_chunk(self):
        raise NotImplementedError('Not implemented yet')

    def _get_files_by_string_matching(
            self, business_id: str, string_match: str,
            app_name: Optional[str] = None,
            allow_reserved_words: bool = False,
    ) -> List[Dict]:
        if not allow_reserved_words:
            if '_date:' in string_match or '_chunk:' in string_match:
                raise ValueError(
                    'Reserved keywords "_date:" and "chunk:" are not allowed in file name'
                )

        apps: List[Dict] = self._get_business_apps(business_id)
        target_files: List[Dict] = []
        for app in apps:
            app_id: str = app['id']

            if app_name is not None:
                if app['name'] != app_name:
                    continue

            files: List[Dict] = self.get_files(business_id=business_id, app_id=app_id)
            for file_metadata in files:
                if string_match in file_metadata['name']:
                    target_files.append(file_metadata)

        return target_files

    def get_files_by_name_prefix(
            self, business_id: str, name_prefix: str,
            app_name: Optional[str] = None,
            allow_reserved_words: bool = False,
    ) -> List[Dict]:
        target_files: List[Dict] = list()
        files: List[Dict] = self._get_files_by_string_matching(
            business_id=business_id,
            string_match=name_prefix,
            app_name=app_name,
            allow_reserved_words=allow_reserved_words,
        )
        for file in files:
            if file['name'].startswith(name_prefix):
                target_files.append(file)
        return target_files

    def get_file_by_name(
            self, business_id: str, file_name: str,
            app_name: Optional[str] = None,
            get_file_object: bool = False,
            force_name: bool = False,
    ) -> Union[Dict, bytes]:
        if not force_name:
            if '_date:' in file_name or '_chunk:' in file_name:
                raise ValueError(
                    'Reserved keywords "_date:" and "chunk:" are not allowed in file name'
                )

        apps: List[Dict] = self._get_business_apps(business_id)
        target_files: List[Dict] = []
        app_id = None
        for app in apps:
            app_id: str = app['id']

            if app_name is not None:
                if app['name'] != app_name:
                    continue

            files: List[Dict] = self.get_files(business_id=business_id, app_id=app_id)
            for file_metadata in files:
                if file_name == file_metadata['name']:
                    target_files.append(file_metadata)
                    break  # there must be only one file with the same name!
            break

        try:
            file: Dict = target_files[0]
        except IndexError:
            raise ValueError(f'File not found | file_name: {file_name}')

        if get_file_object:
            return self._get_file(
                business_id=business_id,
                app_id=app_id,
                file_id=file['id'],
            )
        return file

    def delete_files_by_name_prefix(
            self, business_id: str, name_prefix: str,
            app_name: Optional[str] = None,
    ):
        app: Dict = self._get_app_by_name(business_id=business_id, name=app_name)
        if not app:
            raise ValueError(f'App not found | App name: {app_name}')

        app_id: str = app['id']
        app_name: str = app['name']

        files_metadata: List[Dict] = self.get_files_by_name_prefix(
            business_id=business_id,
            app_name=app_name,
            name_prefix=name_prefix,
        )

        for file_metadata in files_metadata:
            self._delete_file(
                business_id=business_id,
                app_id=app_id,
                file_id=file_metadata['id'],
            )

    def post_object(
            self, business_id: str,
            app_name: str,
            file_name: str,
            object_data: bytes,
            overwrite: bool = True,
            force_name: bool = False,
            content_type: str = 'text/csv',
            date: Optional[dt.date] = dt.date.today(),
    ) -> Dict:
        app: Dict = self._get_app_by_name(business_id=business_id, name=app_name)
        if not app:
            raise ValueError(f'App not found | App name: {app_name}')

        app_id: str = app['id']

        if force_name:
            final_file_name: str = file_name
        else:
            final_file_name: str = self._encode_file_name(
                file_name=file_name, date=date,
            )

        file_metadata = {
            'name': final_file_name,
            'fileName': re.sub('[^0-9a-zA-Z]+', '-', final_file_name).lower(),
            'contentType': content_type,
        }

        if overwrite:
            app: Dict = self._get_app_by_name(business_id=business_id, name=app_name)
            if not app:
                raise ValueError(f'App not found | App name: {app_name}')

            fs: List[Dict] = self.get_files_by_name_prefix(
                business_id=business_id,
                app_name=app_name,
                name_prefix=file_metadata['name'],
                allow_reserved_words=True
            )
            for f in fs:
                self._delete_file(
                    business_id=business_id,
                    app_id=app['id'],
                    file_id=f['id'],
                )

        return self._create_file(
            business_id=business_id,
            app_id=app_id,
            file_metadata=file_metadata,
            file_object=object_data,
        )


class FileMetadataApi(BasicFileMetadataApi, ABC):
    """
    """
    def __init__(self, api_client):
        self.api_client = api_client

    def get_all_files_by_app_name(self, business_id: str, app_name: str) -> List[Dict]:
        """"""
        apps: List[Dict] = self._get_business_apps(business_id)
        files: List[Dict] = []
        for app in apps:
            app_id: str = app['id']

            if app['name'] != app_name:
                continue

            files: List[Dict] = self.get_files(business_id=business_id, app_id=app_id)
            break
        return files

    def get_all_files_by_creation_date(
            self, business_id: str, datetime: dt.datetime,
            app_name: Optional[str] = None,
            force_name: bool = False,
    ) -> List[Dict]:
        """Get all files from a target creation date"""
        raise NotImplementedError

    def get_file_by_creation_date(
            self, business_id: str,
            file_name: str,
            date: dt.date,
            app_name: Optional[str] = None,
            get_file_object: bool = False,
    ) -> Any:
        """Get a specific file from a target creation date"""
        raise NotImplementedError

    def get_last_created_target_file(
            self, business_id: str,
            file_name: str,
            app_name: Optional[str] = None,
    ) -> Dict:
        """Get a specific file from a target creation date"""
        raise NotImplementedError

    def get_all_files_by_date(
            self, business_id: str, date: dt.date,
            app_name: Optional[str] = None,
    ) -> List[Dict]:
        """Get all files from a target date"""
        datetime_str: str = date.isoformat()
        return self._get_files_by_string_matching(
            business_id=business_id,
            string_match=datetime_str,
            app_name=app_name,
        )

    def get_file_by_date(
            self, business_id: str,
            file_name: str,
            date: dt.date,
            app_name: Optional[str] = None,
            get_file_object: bool = False,
    ) -> Any:
        """Get a specific file from a target date"""
        files: List[Dict] = self.get_files_by_name_prefix(
            business_id=business_id,
            app_name=app_name,
            name_prefix=file_name,
        )

        target_files: List[Dict] = [
            file for file in files
            if self._get_file_date(file['name']) == date
        ]

        # Chunks are only allowed for tabular data use get_dataframe() instead!
        len_target_files = len(target_files)
        if len_target_files == 0:
            if get_file_object:  # return Object
                return None
            else:
                return {}
        elif len_target_files >= 1:
            if get_file_object:  # return List[Object]
                try:
                    app: Dict = [
                        app for app in self._get_business_apps(business_id)
                        if app['name'] == app_name
                    ][0]
                except IndexError:
                    raise ValueError(f'App not found | app_name: {app_name}')

                results: List[bytes] = []
                for file in target_files:
                    result = self._get_file(
                        business_id=business_id,
                        app_id=app['id'],
                        file_id=file['id'],
                    )
                    results = results + [result]
                return results
            else:  # return List[Dict]
                return target_files
        else:
            raise ValueError('Multiple files found for the same date')

    def get_files_with_max_date(
            self, business_id: str,
            file_name: str,
            app_name: Optional[str] = None,
            get_file_object: bool = False,
    ) -> Any:
        """Get a specific file with maximum date"""
        app: Dict = self._get_app_by_name(business_id=business_id, name=app_name)
        if not app:
            raise ValueError(f'App not found | App name: {app_name}')

        files: List[Dict] = self.get_files_by_name_prefix(
            business_id=business_id,
            name_prefix=file_name,
            app_name=app_name,
        )
        target_files: List[Dict] = []
        max_date: dt.date = dt.date(2000, 1, 1)
        for file in files:
            try:
                new_date: Optional[dt.date] = self._get_file_date(file['name'])
            except ValueError:
                continue

            if new_date is None:
                continue

            if new_date > max_date:
                target_files = [file]
                max_date = new_date
            elif new_date == max_date:
                target_files = target_files + [file]

        if get_file_object:  # return List[bytes]
            results: List[bytes] = []
            for target_file in target_files:
                result = self._get_file(
                    business_id=business_id,
                    app_id=app['id'],
                    file_id=target_file['id'],
                )
                results = results + [result]
            return results
        else:  # return Dict
            return target_files

    # TODO pending implementation
    def replace_file_name(
            self, business_id: str, app_name: str, old_name: str, new_name: str
    ) -> Dict:
        """
        app: Dict = self._get_app_by_name(business_id=business_id, name=app_name)

        if not app:
            raise ValueError('App name not found')
        app_id: str = app['id']
        app_name: str = app['name']

        file_metadata: Dict = self.get_file_by_name(
            business_id=business_id,
            app_name=app_name,
            file_name=old_name,
        )

        file_object: bytes = self._get_file(
            business_id=business_id,
            app_id=app_id,
            file_id=file_metadata['id'],
        )
        file_metadata['name'] = new_name

        file: Dict = self._create_file(
            business_id=business_id,
            app_id=app_id,
            file=file_object,
            file_metadata=file_metadata,
        )

        self.delete_files_by_name_prefix(
            business_id=business_id,
            app_name=app_name,
            file_name=old_name,
        )
        return file
        """
        raise NotImplementedError

    def get_object(
            self, business_id: str,
            app_name: str,
            file_name: str,
            force_name: bool = False,
    ) -> bytes:
        if force_name:
            return self.get_file_by_name(
                business_id=business_id,
                app_name=app_name,
                file_name=file_name,
                get_file_object=True,
                force_name=True,
            )
        else:
            return self.get_files_with_max_date(
                business_id=business_id,
                file_name=file_name,
                app_name=app_name,
                get_file_object=True,
            )

    def post_dataframe(
            self, business_id: str, app_name: str, file_name: str, df: pd.DataFrame,
            force_name: bool = False, split_by_size: bool = True
    ) -> Union[Dict, List[Dict]]:
        """
        :param business_id:
        :param app_name:
        :param file_name:
        :param df:
        :param force_name: if not forced then we add the date to the file name
        :param split_by_size: if true then we split the file by size and set it in the file name
        """
        if split_by_size:
            len_df: int = len(df)
            bytes_size_df: int = df.memory_usage(deep=True).sum()
            bytes_size_df: int = bytes_size_df / 1024 / 1024
            if bytes_size_df > 5:
                objects: List[Dict] = []
                total_chunks: int = int(bytes_size_df / 5)
                chunk_rows: int = int(len_df / total_chunks)
                for chunk in range(total_chunks + 1):
                    df_temp: pd.DataFrame = df.iloc[chunk*chunk_rows:(chunk+1)*chunk_rows]
                    dataframe_binary: bytes = df_temp.to_csv(index=False).encode('utf-8')

                    if force_name:
                        final_file_name: str = file_name
                    else:
                        final_file_name: str = self._encode_file_name(
                            file_name=file_name, date=dt.date.today(), chunk=chunk
                        )
                    object = self.post_object(
                        business_id=business_id,
                        app_name=app_name,
                        file_name=final_file_name,
                        object_data=dataframe_binary,
                        force_name=True,
                    )
                    objects = objects + [object]
                return objects

        dataframe_binary: bytes = df.to_csv(index=False).encode('utf-8')

        return self.post_object(
            business_id=business_id,
            app_name=app_name,
            file_name=file_name,
            object_data=dataframe_binary,
        )

    def get_dataframe(
            self, business_id: str, app_name: str, file_name: str,
            get_last_date: bool = True, parse_name: bool = True,
            date: Optional[dt.date] = None,
    ) -> pd.DataFrame:
        """
        :param business_id:
        :param app_name:
        :param file_name:
        :param parse_name: if true then we parse the file name to get the date
        :param get_last_date:
        :param date:
        """
        if parse_name:
            df = pd.DataFrame()
            if get_last_date:
                dataset_binary = self.get_files_with_max_date(
                    business_id=business_id,
                    app_name=app_name,
                    file_name=file_name,
                    get_file_object=True,
                )
                for binary in dataset_binary:
                    df = df.append(pd.read_csv(StringIO(binary.decode('utf-8'))))
            elif date:
                dataset_binary = self.get_file_by_date(
                    business_id=business_id,
                    app_name=app_name,
                    file_name=file_name,
                    date=date,
                )
                for binary in dataset_binary:
                    df = df.append(pd.read_csv(StringIO(binary.decode('utf-8'))))
            else:
                dataset_names: List[Dict] = (
                    self.get_files_by_name_prefix(
                        business_id=business_id,
                        app_name=app_name,
                        name_prefix=file_name,
                    )
                )

                for dataset_name in dataset_names:
                    dataset_binary = self.get_object(
                        business_id=business_id,
                        app_name=app_name,
                        file_name=dataset_name['name'],
                        force_name=True,
                    )
                    for binary in dataset_binary:
                        df = df.append(pd.read_csv(StringIO(binary.decode('utf-8'))))
        else:
            dataset_binary: bytes = self.get_file_by_name(
                business_id=business_id,
                app_name=app_name,
                file_name=file_name,
                get_file_object=True,
            )
            d = StringIO(dataset_binary.decode('utf-8'))
            df = pd.read_csv(d)

        return df.reset_index(drop=True)

    def post_ai_model(
            self, business_id: str, app_name: str, model_name: str, model: Callable,
    ) -> Dict:
        """
        :param business_id:
        :param app_name:
        :param model_name:
        :param model:
        """
        model_binary: bytes = pickle.dumps(model)

        return self.post_object(
            business_id=business_id,
            app_name=app_name,
            file_name=model_name,
            object_data=model_binary,
        )

    def get_ai_model(
            self, business_id: str, app_name: str, model_name: str,
    ) -> Any:
        """
        :param business_id:
        :param app_name:
        :param model_name:
        """
        model_binary: bytes = self.get_object(
            business_id=business_id,
            app_name=app_name,
            file_name=model_name,
        )

        assert len(model_binary) == 1
        model_binary = model_binary[0]

        return pickle.loads(model_binary)



"""
To create a report

r = {
    'business_id': business_id,
    'app_id': app_id,
    'report_metadata': {
        'title': 'test',
        'order': 0,
        'grid': '1, 1',
        'path': 'test',
        'reportType': 'LINECHART',
        'chartData': '[{"name": "Matcha Latte", "value": 78}, {"name": "Milk Tea", "value": 17}, {"name": "Cheese Cocoa", "value": 18}, {"name": "Walnut Brownie", "value": 9}]'
    }
}
s.report.create_report(**r)

"""
from os import getenv
from typing import Dict, List
import unittest

import datetime as dt

import shimoku_api_python as shimoku
from shimoku_api_python.exceptions import ApiClientError


api_key: str = getenv('API_TOKEN')
universe_id: str = getenv('UNIVERSE_ID')
business_id: str = getenv('BUSINESS_ID')
app_id: str = getenv('APP_ID')
report_id: str = getenv('REPORT_ID')
environment: str = getenv('ENVIRONMENT')
report_element: Dict[str, str] = dict(
    business_id=business_id,
    app_id=app_id,
    report_id=report_id
)


config = {
    'access_token': api_key,
}

s = shimoku.Client(
    config=config,
    universe_id=universe_id,
    environment=environment,
)


def test_get_report():
    report: Dict = s.report.get_report(**report_element)
    assert report


def test_update_report():
    """Set the updatedAt field of an report to '2000-01-01'
    Then revert the updatedAt to its original value
    """
    report: Dict = s.report.get_report(**report_element)
    old_val: str = report['reportType']

    val: str = 'BARCHART'
    assert old_val != val

    report_data: Dict = {'reportType': val}
    s.report.update_report(
        report_metadata=report_data,
        **report_element,
    )

    report_updated: Dict = s.report.get_report(**report_element)

    assert report_updated['reportType'] == val

    #########
    # Revert the change
    #########

    report_data: Dict = {'reportType': old_val}
    s.report.update_report(
        report_metadata=report_data,
        **report_element,
    )

    report_updated: Dict = s.report.get_report(**report_element)

    assert report_updated['reportType'] == old_val


def test_create_and_delete_report():
    report_metadata = {
        'title': 'test',
        'order': 0,
        'grid': '1, 1',
        'path': 'test',
        'reportType': 'LINECHART',
    }
    new_report: Dict = (
        s.report.create_report(
            business_id=business_id,
            app_id=app_id,
            report_metadata=report_metadata,
        )
    )
    new_report_id: str = new_report['id']

    report: Dict = s.report.get_report(
        business_id=business_id,
        app_id=app_id,
        report_id=new_report_id,
    )

    assert new_report == {
        k: v
        for k, v in report.items()
        if k in [
            'id', 'appId', 'reportDataSets',
            'path', 'pathOrder', 'grid',
            'createdAt', 'order', 'reportType',
            'dataFields', 'title', 'sizeRows',
            'sizeColumns', 'sizePadding', 'smartFilters',
            'subscribe', 'updatedAt', 'bentobox',
            'hidePath', 'isDisabled',
            'properties', '__typename',
        ]
    }

    s.report.delete_report(
        business_id=business_id,
        app_id=app_id,
        report_id=new_report_id,
    )

    # Check it does not exists anymore
    class MyTestCase(unittest.TestCase):
        def check_report_not_exists(self):
            with self.assertRaises(ApiClientError):
                s.report.get_report(
                    business_id=business_id,
                    app_id=app_id,
                    report_id=new_report_id,
                )

    t = MyTestCase()
    t.check_report_not_exists()


def test_get_report_data():
    data: List[Dict] = s.report.get_report_data(**report_element)
    assert data
    assert len(data[0]) > 0


def test_get_reports_in_same_app():
    reports: List[str] = s.report.get_reports_in_same_app(**report_element)
    assert reports
    assert len(reports) > 1


def test_get_reports_in_same_path():
    reports: List[str] = s.report.get_reports_in_same_path(**report_element)
    assert reports
    assert len(reports) > 1


def test_get_report_last_update():
    last_update: dt.datetime = s.report.get_report_last_update(**report_element)


def test_get_report_by_title():
    title: str = 'test'
    reports: Dict = (
        s.report.get_reports_by_title(
            business_id=business_id,
            app_id=app_id,
            title=title,
        )
    )
    assert reports
    assert all([r['title'] == title for r in reports])


def test_get_report_by_path():
    path: str = 'test'
    reports: Dict = (
        s.report.get_reports_by_path(
            business_id=business_id,
            app_id=app_id,
            path=path,
        )
    )
    assert reports
    assert all([r['path'] == path for r in reports])


def test_get_report_by_chart_type():
    report_type: str = 'LINECHART'
    reports: List[Dict] = (
        s.report.get_reports_by_chart_type(
            business_id=business_id,
            app_id=app_id,
            chart_type=report_type,
        )
    )
    assert reports
    assert all([r['reportType'] == report_type for r in reports])


def test_get_report_by_grid_position():
    row: int = 1
    column: int = 1
    grid = f'{row}, {column}'
    reports: List[Dict] = (
        s.report.get_reports_by_grid_position(
            business_id=business_id,
            app_id=app_id,
            row=row, column=column,
        )
    )
    assert reports
    assert all([r['grid'].strip() == grid for r in reports])


# TODO
def test_change_report_grid_position():
    # s.report.change_report_grid_position()
    raise NotImplementedError


def test_fetch_filter_report():
    # s.report.fetch_filter_report()
    raise NotImplementedError


def test_add_report_to_filter():
    # s.report.add_report_to_filter()
    raise NotImplementedError


def test_set_filter_to_reports():
    # s.report.set_filter_to_reports()
    raise NotImplementedError


def test_remove_filter_for_report():
    # s.report.remove_filter_for_report()
    raise NotImplementedError


def test_hide_report():
    s.report.hide_report(**report_element)
    assert s.report.get_report(**report_element)['isDisabled']


def test_unhide_report():
    s.report.unhide_report(**report_element)
    assert not s.report.get_report(**report_element)['isDisabled']


test_get_report()
test_create_and_delete_report()
test_update_report()
test_get_reports_in_same_path()
test_get_report_data()

test_get_report_by_title()
test_get_report_by_path()
test_get_report_by_chart_type()
test_get_report_by_grid_position()
test_hide_report()
test_unhide_report()

# TODO
"""
test_change_report_grid_position()
test_get_filter_report()
test_add_report_to_filter()
test_remove_filter_for_report()
test_set_filter_to_reports()
test_fetch_filter_report()
"""
